"""Experiment module on adult dataset.
It creates the directory results/<generator>/<k> to save three dictionaries of metrics:
    - utility.pickle
    - mia.pickle
    - aia.pickle"""

from pathlib import Path
import pickle
import os

from ..fetch_data import adult
from ..generator import identity
from ..generator.adult import ctgan
from ..predictor.adult import AdultNN
from ..aia import Aia
from ..mia import Mia, random_fusion
from ..metrics import balanced_accuracy

def _aia(dadata):
    """Run attribute inference attack on a dataset containing preditions and return attack results.

    :param data: Dataset with prediction.
    :type data: pandas.dataframe
    :return: Attack metrics.
    :rtype: dictionary
    """
    data = dadata.load()
    skf = StratifiedKFold(data)
    data_train = Dataset()
    data_tes = Dataset()
    for train,test in skf:
        data_train.update(data.iloc[train])
        data_test.update(data.iloc[test])
        aia = Aia()
        aia.fit(data_train)
        pred = aia.predict(data_test)
        for attrib in ["sex", "race"]:
            metrics[attrib] = {}
            for labtype in ["hard","soft"]:
                metrics[attrib][labtype] = {}
                for metric in metric_lst:
                    value = balanced_accuracy(pred, attrib, f"{attrib}_{labtype}")
                    metrics[attrib][labtype][metric] += [value]

    return metrics

def _mia(dadata):
    """Run membership inference attack on a dataset containing point members and non members with associated logits.

    :para dadata: Dataset with memeber and non memebers labeled.
    :type dadata: fetch_data.Dataset
    :return: Attack metrics.
    :rtype: dictionary
    """

    data = dadata.load()
    skf = StratifiedKFold(data,data["member"])
    data_train = Dataset()
    data_test = Dataset()

    metrics = {"ba":[]}
    for train,test in skf:
        data_train.update(data.iloc(train))
        data_train.update(data.iloc(test))
        mia = Mia()
        mia.fit(data_train)
        pred = mia.predict(data_test)
        metrcis["ba"] += [balanced_accuracy(pred, "member", "mia")]

    return metrics

def _utility(dadata):
    """Compute utility metrics.

    :param dadata: Dataset containing prediction as a column named "hard" and true label (PINCP).
    :type dadate: fetch_data.Dataset
    :return: Utility metrics.
    :rtype: Dictionary
    """

    metrics = {"ba":balanced_accuracy(pred, "PINCP", "hard")}
    return metrics



def _loop(generator,k):
    """Run in parallel the cross validations steps.
    
    :param generate: A function to generate synthetic data. It takes as an input a dataset and return synthetic data with as many datarecord as in the input.
    :type generate: function
    :param k: Cross validation step.
    :type k: int
    """
    data = adult.load(["sex","race"],k=k)

    synthetic = generator(data["train"])

    adultNN = AdultNN()
    adultNN.fit(synthetic)
    pred = adultNN.predict(data["test"])

    utility_metrics = _utility(pred)
    
    aia_metrics = _aia(pred)

    mia_data = random_fusion(data["train"],data["test"])
    pred = adultNN.predict(mia_data)
    mia_metrics = _mia(pred)

    path = Path("results",generator,k)
    os.makedirs(path,exist_ok=True)

    with open(Path(path,"utility.pickle"), 'wb') as f:
        pickle.dump(aia_metrics, f)

    with open(Path(path,"aia.pickle"), 'wb') as f:
        pickle.dump(utility_metrics, f)

    with open(Path(path,"mia.pickle"), 'wb') as f:
        pickle.dump(mia_metrics, f)



if __name__=="__main__":
    import sys
    generator_names = {"id":identity,"ctgan":ctgan}
    error_message = f"""Invalid command synthax.
Correct synthax is:
python -m synthetic.experiments.adult <generator_name> <k>
where 
    <generator_name> is on of the following:
    {list(generator_names.keys())}

    <k> is the cross validation step in:
    0,1,2,3,4
    """
    try:
        user_argument = sys.argv[1]
    except IndexError:
        raise ValueError(error_message)
        sys.exit()

    try:
        k = int(sys.argv[2])
    except IndexError:
        raise ValueError(error_message)
        sys.exit()

    if not(user_argument in generator_names.keys()):
        raise ValueError(error_message)
        sys.exit()

    if not(k in [0,1,2,3,4]):
        raise ValueError(error_message)
        sys.exit()

    generator = generator_names[user_argument]
    _loop(generator,k=k)



